#include "gpioexceptions.h"
#include "bbb_gpio.h"
#include <iostream>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>

CGPIO::CGPIO(gpio_ports port,int pin)
{
  int fp; 
  struct stat sb;
  std::stringstream text,gpio_file;

  /* try to export the desired pin */
  if((fp=open("/sys/class/gpio/export",O_WRONLY))==-1)
  {  
    /* handle exceptions */
    throw CGPIOException(_HERE_,"Impossible to export the desired port pin");
  }
  // set the file pointer to the beginning
  lseek(fp,0,SEEK_SET);
  // select the desired GPIO index
  if(pin < 0 || pin > 31)
  {
    /* handle exceptions */
    close(fp);
    throw CGPIOException(_HERE_,"Invalid pin identifier");
  }
  this->gpio_id=(int)port+pin;
  gpio_file << "/sys/class/gpio/gpio" << this->gpio_id;
  /* check if the file already exists */
  if(stat(gpio_file.str().c_str(),&sb)!=0)
  {
    text << this->gpio_id;
    if(write(fp,text.str().c_str(),text.str().size())!=(int)text.str().size())
    {
      /* handle exceptions */
      close(fp);
      throw CGPIOException(_HERE_,"Error while writing to the export file");
    }
  }
  // close the export file
  close(fp);
  // initialize internal parameters
  this->direction=input;
  this->edge=edge_none;
  this->event_server=CEventServer::instance();
  text.str(std::string(""));
  text << "pin_" << this->gpio_id << "_finish_thread_event";
  this->finish_thread_event_id=text.str();
  this->event_server->create_event(this->finish_thread_event_id);
  this->thread_server=CThreadServer::instance();
  text.str(std::string(""));
  text << "pin_" << this->gpio_id << "_thread";
  this->edge_thread_id=text.str();
  this->thread_server->create_thread(this->edge_thread_id);
  this->thread_server->attach_thread(this->edge_thread_id,this->edge_thread,this);
  this->input_fd=-1;
}

void *CGPIO::edge_thread(void *param)
{
  CGPIO *gpio=(CGPIO *)param;
  unsigned char buffer[1024];
  int error,max_fd=0;
  fd_set wait_set;
  bool end=false;
  timeval time;

  while(!end)
  {
    if(gpio->event_server->event_is_set(gpio->finish_thread_event_id))
      end=true;
    else
    {
      time.tv_sec=1;
      time.tv_usec=0;
      FD_ZERO(&wait_set);
      FD_SET(gpio->input_fd,&wait_set);
      max_fd=gpio->input_fd;
      error=select(max_fd+1,NULL,NULL,&wait_set,&time);
      if(error==-1)
      {
        if(errno!=EINTR)
        {
          /* handle exceptions */
          throw CGPIOException(_HERE_,"Unexpected error while waiting the activation of one the events");
        }
      }
      else if(error==0)
      {
        /* timeout, do nothing */
      }
      else
      {
        lseek(gpio->input_fd,0,SEEK_SET);
        read(gpio->input_fd,buffer,1024);
        if(FD_ISSET(gpio->input_fd,&wait_set))
          if(!gpio->event_server->event_is_set(gpio->edge_event_id))
            gpio->event_server->set_event(gpio->edge_event_id);
      }
    }
  }

  pthread_exit(NULL);
}

void CGPIO::set_direction(dir_t direction)
{
  int fp;
  struct stat sb;
  std::stringstream text;

  text << "/sys/devices/virtual/gpio/gpio" << this->gpio_id << "/direction";
  //Open the GPIO's sysfs file for writing
  while(stat(text.str().c_str(),&sb)!=0 || (sb.st_mode&(S_IWUSR | S_IWGRP))!=(S_IWUSR | S_IWGRP))
    usleep(100000);
  if((fp=open(text.str().c_str(), O_WRONLY)) == -1)
  {
    /* handle exceptions */
    throw CGPIOException(_HERE_,"Impossible to open the direction file");
  }
  //Set pointer to begining of the file
  lseek(fp,0,SEEK_SET);
  //Write our value of "out" to the file
  text.str(std::string());
  if(direction==input)
    text << "in";
  else
  {
    text << "out";
    // finish the thread if it is active
    if(this->thread_server->get_thread_state(this->edge_thread_id)==starting || 
       this->thread_server->get_thread_state(this->edge_thread_id)==active)
    {
      this->event_server->set_event(this->finish_thread_event_id);
      this->thread_server->end_thread(this->edge_thread_id);
      this->event_server->reset_event(this->finish_thread_event_id);
    }
    // close the input file
    close(this->input_fd);
    this->input_fd=-1;
  }
  if(write(fp,text.str().c_str(),text.str().size())!=(int)text.str().size())
  {
    close(fp);
    throw CGPIOException(_HERE_,"Error while writing to the direction file");
  }
  close(fp);

  this->direction=direction;
}

void CGPIO::set_value(bool value)
{
  int fp;
  struct stat sb;
  std::stringstream text;

  if(this->direction==output)
  {
    //Open the GPIO's sysfs file for reading and writing
    text << "/sys/devices/virtual/gpio/gpio" << this->gpio_id << "/value";
    while(stat(text.str().c_str(),&sb)!=0 || (sb.st_mode&(S_IWUSR | S_IWGRP))!=(S_IWUSR | S_IWGRP))
      usleep(100000);
    if((fp=open(text.str().c_str(),O_RDWR)) == -1)
    {
      /* handle exceptions */
      throw CGPIOException(_HERE_,"Impossible to open the value file");
    }
    //Set pointer to begining of the file
    lseek(fp,0,SEEK_SET);
    //Write our value of "1" to the file 
    text.str(std::string());
    if(value)
      text << 1;
    else
      text << 0;
    if(write(fp,text.str().c_str(),text.str().size())!=(int)text.str().size())
    {
      close(fp);
      throw CGPIOException(_HERE_,"Error while writing to the value file");
    }
    close(fp);
  }
}

bool CGPIO::get_value(void)
{
  int fp;
  struct stat sb;
  unsigned char value;
  std::stringstream text;

  if(this->direction==input)
  {
    //Open the GPIO's sysfs file for reading and writing
    text << "/sys/devices/virtual/gpio/gpio" << this->gpio_id << "/value";
    while(stat(text.str().c_str(),&sb)!=0 || (sb.st_mode&(S_IWUSR | S_IWGRP))!=(S_IWUSR | S_IWGRP))
      usleep(100000);
    if((fp=open(text.str().c_str(),O_RDWR)) == -1)
    {
      /* handle exceptions */
      throw CGPIOException(_HERE_,"Impossible to open the value file");
    }
    //Set pointer to begining of the file
    lseek(fp,0,SEEK_SET);
    if(read(fp,&value,1)!=1)
    {
      close(fp);
      throw CGPIOException(_HERE_,"Error while writing to the value file");
    }
    close(fp);
    if(value=='0') return false;
    else return true;
  }
  return false;
}

std::string CGPIO::set_active_edge(edge_t edge)
{
  int fp; 
  struct stat sb;
  std::stringstream text;
  unsigned char buffer[1024];

  if(this->direction==input)
  {
    //Open the GPIO's sysfs file for reading and writing
    text << "/sys/devices/virtual/gpio/gpio" << this->gpio_id << "/edge";
    while(stat(text.str().c_str(),&sb)!=0 || (sb.st_mode&(S_IWUSR | S_IWGRP))!=(S_IWUSR | S_IWGRP))
      usleep(100000);
    if((fp=open(text.str().c_str(),O_WRONLY)) == -1)
    {
      /* handle exceptions */
      throw CGPIOException(_HERE_,"Impossible to open the edge file");
    }
    //Set pointer to begining of the file
    lseek(fp,0,SEEK_SET);
    //Write our value of "1" to the file 
    text.str(std::string());
    switch(edge)
    {
      case edge_none: text << "none";
                      break;
      case edge_rising: text << "rising";
                        break;
      case edge_falling: text << "falling";
                         break;
      case edge_both: text << "both";
                       break;
    }
    if(write(fp,text.str().c_str(),text.str().size())!=(int)text.str().size())
    {
      close(fp);
      throw CGPIOException(_HERE_,"Error while writing to the edge file");
    }

    close(fp);

    this->edge=edge;

    // open the input file
    text.str(std::string(""));
    text << "/sys/devices/virtual/gpio/gpio" << this->gpio_id << "/value";
    while(stat(text.str().c_str(),&sb)!=0 || (sb.st_mode&(S_IWUSR | S_IWGRP))!=(S_IWUSR | S_IWGRP))
      usleep(100000);
    if((this->input_fd=open(text.str().c_str(),O_RDONLY|O_NONBLOCK)) == -1)
    {
      /* handle exceptions */
      throw CGPIOException(_HERE_,"Impossible to open the value file");
    }
    //Set pointer to begining of the file
    lseek(this->input_fd,0,SEEK_SET);
    read(this->input_fd,buffer,1024);
    if(this->edge!=edge_none)
    {
      if(this->thread_server->get_thread_state(this->edge_thread_id)==attached)
      {
        // start the waiting thread
        this->thread_server->start_thread(this->edge_thread_id);
        // create the edge event
        text.str(std::string(""));
        text << "pin_" << this->gpio_id << "_event";
        this->edge_event_id=text.str();
        this->event_server->create_event(this->edge_event_id);
      }
    }
    else
    {
      // stop the thread if it is running
      if(this->thread_server->get_thread_state(this->edge_thread_id)==starting ||
         this->thread_server->get_thread_state(this->edge_thread_id)==active)
      {
        this->event_server->set_event(this->finish_thread_event_id);
        this->thread_server->end_thread(this->edge_thread_id);
      }
      if(this->edge_event_id!="")
        this->event_server->delete_event(this->edge_event_id);
      this->edge_event_id="";
    }

    return this->edge_event_id;
  }
  else 
    return std::string("");
}

CGPIO::~CGPIO()
{
  int fp; 
  std::stringstream text;

  // finish the thread if necessary
  if(this->thread_server->get_thread_state(this->edge_thread_id)==starting || 
     this->thread_server->get_thread_state(this->edge_thread_id)==active)
  {
    this->event_server->set_event(this->finish_thread_event_id);
    this->thread_server->end_thread(this->edge_thread_id);
    this->event_server->delete_event(this->edge_event_id);
    this->edge_event_id="";
  }
  this->event_server->delete_event(this->finish_thread_event_id);
  this->finish_thread_event_id="";
  this->thread_server->detach_thread(this->edge_thread_id);
  this->thread_server->delete_thread(this->edge_thread_id);
  this->edge_thread_id="";
  
  // cloe the input file descriptor if necessary
  if(this->input_fd!=-1)
  {
    close(this->input_fd);
    this->input_fd=-1;
  }

  /* try to unexport the desired pin */
  if((fp=open("/sys/class/gpio/unexport",O_WRONLY))==-1)
  {  
    /* handle exceptions */
    throw CGPIOException(_HERE_,"Impossible to unexport the desired port pin");
  }
  // set the file pointer to the beginning
  lseek(fp,0,SEEK_SET);
  text << this->gpio_id;
  if(write(fp,text.str().c_str(),text.str().size())!=(int)text.str().size())
  {
    /* handle exceptions */
    close(fp);
    throw CGPIOException(_HERE_,"Error while writing to the unexport file");
  }
  // close the export file
  close(fp);
}
