#ifndef GPIO_EXCEPTIONS
#define GPIO_EXCEPTIONS

#include "exceptions.h"

/**
 * \brief Exception class for the CGPIO class
 *  
 * This class is used as the main exception to be thrown from the CGPIO class.
 * It inherits from the basic CException class from iriutils and preappends
 * some additional information to the standard error message.
 */
class CGPIOException : public CException
{
  public:
    /**
     * \brief Constructor
     *
     * The constructor calls the base class constructor to add the general
     * exception identifier and then adds the class identifier string 
     * "[CGPIO class]" and the supplied error message. 
     *
     * So, the total exception message will look like this:
     *
     * \verbatim
     * [Exception caught] - <where>
     * [CGPIO class] - <error message> 
     * \endverbatim
     *
     * \param where a null terminated string with the information about the name
     *              of the function, the source code filename and the line where
     *              the exception was generated. This string must be generated 
     *              by the _HERE_ macro.
     *
     * \param error_msg a null terminated string that contains the error message.
     *                  This string may have any valid character and there is no 
     *                  limit on its length.
     */
    CGPIOException(const std::string& where, const std::string& error_msg);
};

#endif
